#!/usr/bin/perl
#
# This rebuilds your indexes.  Those Mini-ITX machines are slow,
# and I have thousands of MP3s.

use strict;
use Storable;

my $dir = "/mp3/";
chdir $dir;

# Format:
#
#  SEARCHABLE
#  actual         (path to file)
#  display        (with leading numbers removed (pattern below))
#
# Example:
#
#  HOLE LIVETHROUGHTHIS LIVETHROUGHTHIS
#  Hole/1994-Live Through%20This/Live Through This.m3u
#  Hole/Live Through This/Live Through This
#  HOLE LIVETHROUGHTHIS MISSWORLD
#  Hole/1994-Live Through%20This/02 - Miss World.mp3
#  Hole/Live Through This/Miss World
#  

my @all;

my $lastdir = "";
my @m3u;
my @mp3;

my $out = sub {
    return unless @m3u || @mp3;
    push @all, @m3u, @mp3;
    @m3u = ();
    @mp3 = ();
};

foreach my $file (sort `find .`)
{
    chomp $file;
    $file =~ s!^./!!;
    next unless $file =~ /\.(m3u|mp3)$/i;
    my $type = lc($1);

    my $actual = $file;

    my $display = $actual;
    $display =~ s/(\.wav)?\.(m3u|mp3)$//i;
    $display =~ s/%([0-9a-f]{2,2})/chr hex $1/eg;
    $display =~ s/_/ /g;
    $display =~ s!/\d+((\s*[\.\-]\s*)|\s+)!/!g;

    my $searchable = uc($display);
    $searchable =~ s/[^\w\/]//g;
    $searchable =~ s!/! !g;
    
    $actual =~ m!(.+/)!;
    my $dir = $1;

    if ($dir ne $lastdir) {
	$out->();
	$lastdir = $dir;
    }

    if ($type eq "mp3") { push @mp3, [ $searchable, $actual, $display ] }
    elsif ($type eq "m3u") { push @m3u, [ $searchable, $actual, $display ] }
}
$out->();

my $map = {
    '1' => "1 ",
    '2' => "2ABC",
    '3' => "3DEF",
    '4' => "4GHI",
    '5' => "5JKL",
    '6' => "6MNO",
    '7' => "7PQRS",
    '8' => "8TUV",
    '9' => "9WXYZ",
};

my %count;
my $i = 0;
for my $va (1..9) {
    for my $vb (1..9) {
	$i++;
	printf "Indexing... %0.1f%%\n", 100*$i/81;
	for my $vc (1..9) {
            my $matches = 0;
            my %dir = ('' => { 'name' => '(root)', 'open' => 1 });
	    my $regexp = join('', map { "[$map->{$_}]" } ($va, $vb, $vc));

	    foreach my $it (@all) {
                # $it = [ search, actual, display ]
		next unless $it->[0] =~ /$regexp/;

                $matches++;
                my @spart = split(/\s+/, $it->[0]);
                my @dpart = split(/\//, $it->[2]);
                my $parts = @dpart;

                my $dirkey = join("/", @dpart[0..$parts-2]);
                unless ($dir{$dirkey}) {
                    for (my $i=0; $i < $parts-1; $i++) {
                        $dirkey = join("/", @dpart[0..$i]);
                        unless ($dir{$dirkey}) {
                            my $sdir = uc($dpart[$i]);
                            $sdir =~ s/[^\w]//g;
                            $dir{$dirkey} = {
                                'name' => $dpart[$i],
                                'search' => $sdir,
                                'children' => [],
                                'items' => [],
                            };
                            my $parent = join("/", @dpart[0..$i-1]);
                            push @{$dir{$parent}->{'children'}}, $dirkey;
                        }
                    }
                }

                # see how deep we matched, to decide how much to open
                # up tree
                for (my $i=$parts-1; $i >= 0; $i--) {
                    my $spart = $spart[$i];
                    if ($spart =~ /$regexp/) {
                        # open up all the way to the top
                        for (my $j=$i; $j >= 0; $j--) {
                            my $dirkey = join('/', @dpart[0..$j]);
                            if (ref $dir{$dirkey} eq "HASH") {
                                $dir{$dirkey}->{'open'} = 1;
                            }
                        }
                        # end the loop:
                        $i = -1;
                    }
                }

                my $type = "song";
                if ($it->[1] =~ /\.m3u/) { $type = "playlist"; }
                push @{$dir{$dirkey}->{'items'}}, [ $it->[1], $dpart[$parts-1], $type, $it->[0] ];
            }
            
            $dir{''}->{'matches'} = $matches;
	    store \%dir, ".enty.$va$vb$vc";
	}
    }
}

