package Enty;

use strict;
use Storable;
use Time::HiRes qw(gettimeofday);

use Qt;
use Qt::isa qw(Qt::Widget);

use Qt::slots
    'gotIR' => [],
    'checkBlanking' => [],
    'connectBird' => [],
    'birdError' => ['int'];

use Qt::attributes qw(
		      bird_s
		      );

use NumberLetter;
use EBoxItem;
use HelpPad;

# what number keys have been pressed
my @keylist; 

# screen blanking & dup prevention
my $lastpress;    # seconds
my $lastpress_ms; # milliseconds
my $lastnum;      # number key pressed
my $screen_blanked = 0;

# widget pointers
my $topbox;
my $list;
my $helper;
my $help_on = 0;
my $list_on = 0;

# current search results
my $dir;  # "foo/bar" => { 'open' => 1, 'items' => [...] }

# if true, we just filter $dir, rather than reloading from stored base on disk.
# however, if they press backspace, we'll have to start over from the 3 digit
# base on disk and filter down to wherever we're at.
my $keep_filtering = 0;

sub NEW {
    shift->SUPER::NEW(undef, undef, &Qt::WStyle_Customize | &Qt::WStyle_NoBorder);

    this->setMinimumSize(800,600);
    this->setMaximumSize(800,600);

    connectBird();

    my $grid = Qt::GridLayout(this, 2, 1, 20);

    $topbox = Qt::HBoxLayout;
    $topbox->addStrut(80);

    $grid->addLayout($topbox, 0, 0);

    $list = Qt::ListBox(this, "list");
    $list->setFont(Qt::Font("Helvetica", 16, &Qt::Font::Bold));
    $grid->addWidget($list, 1, 0);
    $list->hide;

    $helper = HelpPad(this);
    $helper->setGeometry(240, 140, 320, 320);
    setHelp(1);

    this->setGeometry(0, 0, 800, 600);
    setBackgroundColor(Qt::Color(0, 0, 0));
    updateChoices();

    # initialize the current time
    ($lastpress, $lastpress_ms) = gettimeofday;

    # every 10 seconds, check to see if it's been 20 seconds
    # since an IR event and blank the screen if needed
    $lastpress = time();
    my $timer = Qt::Timer(this, "periodic");
    this->connect($timer, SIGNAL("timeout()"), SLOT("checkBlanking()"));
    $timer->start(10_000);
}

sub checkBlanking
{
    my $now = time;
    my $blank_in = $lastpress + 20 - $now;
    return if $blank_in > 0;
    return if $screen_blanked;
    setBlanking(1);
}

sub addNumber
{
    my $num = shift;
    my $nl = NumberLetter($num, this);
    push @keylist, $nl;
    $topbox->addWidget($nl);
    $nl->show;

    # updateChoices() might take a second, so we want to update the UI
    # as soon as possible
    Qt::app->processEvents();

    updateChoices();
}

sub setBlanking
{
    my $blank = shift;
    return if $screen_blanked == $blank;
    $screen_blanked = $blank;
    
    if ($blank) {
	foreach (@keylist) { $_->hide; }
	$list->hide;
	$helper->hide;
    } else {
	foreach (@keylist) { $_->show; }
	$list->show if $list_on;
	$helper->show if $help_on;
    }
}

sub setHelp
{
    my $on = shift;
    return if $help_on == $on;
    $help_on = $on;

    if ($on) {
	$helper->show;
    } else {
	$helper->hide;
    }
}

sub setListOn
{
    my $on = shift;
    return if $list_on == $on;
    $list_on = $on;

    if ($on) {
	$list->show;
    } else {
	$list->hide;
    }
}

sub filterChildren
{
    my $parent = shift;
    my $re = shift;

    my $d = $dir->{$parent};
    my $total = 0;  # total matches in all children
    my $needopen = 0;

    # count matching items
    @{$d->{'items'}} = grep { $_->[3] =~ /$re/ } @{$d->{'items'}};
    $total = @{$d->{'items'}};
    
    # see if any match in just last component
    my $finalre = qr/\s+\S*$re\S*$/;
    foreach (@{$d->{'items'}}) {
        $needopen = 1 if $_->[3] =~ /$finalre/;
    }

    # include count of subfolders
    foreach my $ckey (@{$d->{'children'}}) {
        next unless $dir->{$ckey};
        my ($sub_total, $sub_open) = filterChildren($ckey, $re);
        $total += $sub_total;
        $needopen = 1 if $sub_open;
    }

    if ($total) {
        $d->{'open'} = $needopen || $d->{'search'} =~ /$re/;
    } else {
        delete $dir->{$parent};
    }

    return wantarray ? ($total, $needopen) : $total;
}

sub updateChoices
{
    $list->clear();

    if (@keylist == 1) {
        # we might have gotten too wide from two many numbers at top
        this->setGeometry(0, 0, 800, 600);
    }

    if (@keylist < 3) { setListOn(0); return; }

    # if 3 letters, we already have the data structure
    my $base = join('', map { $_->number() } @keylist[0..2]);
    my $matches;
    
    unless ($keep_filtering) {
        $dir = retrieve "/mp3/.enty.$base";
        $keep_filtering = 1;
    }

    # if more than 3, we need to filter our base down
    if (@keylist > 3) {
        my $regexp;
        foreach (@keylist) {
            $regexp .= "[" . $_->letters() . "]";
        }
        $regexp = qr/$regexp/;
        $matches = filterChildren('', $regexp);
    } else {
        # with three presses, conditionally show results, if small.
        # else, wait for 4th press.
        $matches = $dir->{''}->{'matches'};
        return if $matches > 500;
    }

    print STDERR "Matches: $matches\n";

    setListOn(1);
    updateListWidget();
}

sub updateListWidget
{
    $list->clear;
    addListChildrenOf("", 0);
}

sub addListChildrenOf
{
    my $pkey = shift;
    my $indent = shift;

    return unless $dir->{$pkey};
    my $d = $dir->{$pkey};

    my $addItem = sub {
	my ($indent, $type, $display, $file) = @_;
	my $string = "  "x$indent . "[$type] $display";
	#print $string, "\n";
	#$list->insertItem($string, -1);
	$list->insertItem(EBoxItem($indent, $type, $display, $file));
    };

    # show its files
    foreach my $i (@{$d->{'items'}}) {
	my $type = $i->[2];
	$addItem->($indent, $type, $i->[1], $i->[0]);
    }
	
    # show its sub-folders
    foreach my $ckey (@{$d->{'children'}}) {
	my $cd = $dir->{$ckey};
        next unless $cd;
	if ($cd->{'open'}) {
	    $addItem->($indent, "diropen", $cd->{'name'}, $ckey);
	    addListChildrenOf($ckey, $indent+1);
	} else {
	    $addItem->($indent, "dirclosed", $cd->{'name'}, $ckey);
	}
    }

}

sub clearAll
{
    while (@keylist) {
	my $last = pop @keylist;
	$last->hide;
	$last->destroy;
    }
    $keep_filtering = 0;
    updateChoices();
    setHelp(1);
    setListOn(0);
}

sub clearOne
{
    return unless @keylist;
    my $last = pop @keylist;
    $last->hide;
    $last->destroy;
    $keep_filtering = 0;  # force broaden of possibilities
    updateChoices();
    setHelp(1);
}

sub connectBird 
{
    print STDERR "Connect...\n";
    bird_s = Qt::Socket();
    bird_s->connectToHost("localhost", 9034);
    this->connect(bird_s, SIGNAL("error(int)"), SLOT('birdError(int)'));
    this->connect(bird_s, SIGNAL("readyRead()"), SLOT('gotIR()'));
    
}

sub birdError { 
    print STDERR "Unable to connect to IR receiver daemon.\n";
    Qt::app->quit(1);
}

sub gotIR
{
    my $i = bird_s->bytesAvailable();
    my $buf = " "x255;
    my $l = bird_s->readBlock($buf, 255);
    $buf =~ s/\n.*//;
    return unless $buf =~ /^code (\S+)\s*$/;
    my $code = $1;

    my ($now, $now_ms) = gettimeofday;
    my $diff_ms = ($now - $lastpress) * 1_000_000 + ($now_ms - $lastpress_ms);
    ($lastpress, $lastpress_ms) = ($now, $now_ms);

    setBlanking(0) if $screen_blanked;

    my $logkey = {
	'414000000000' => 'clear_one', # 1
	'417000000000' => '2',
	'416000000000' => '3',
	'411000000000' => '4',
	'410000000000' => '5',
	'413000000000' => '6',
	'412000000000' => '7',
	'41d000000000' => '8',
	'41c000000000' => '9',
	'415000000000' => '0',
	'424000000000' => 'clear_all',
	'420000000000' => 'play',
	'423000000000' => 'stop',
	'43c000000000' => 'pause',
	'575000000000' => 'move_up',  # channel up
	'574000000000' => 'move_down', # channel down
	'545000000000' => 'page_up',  # vol_up
	'544000000000' => 'page_down',  # vol_down
	'421000000000' => 'fast_forward',
	'427000000000' => 'rewind',
	'408000000000' => 'select',  # menu
	'41a000000000' => 'status',
    }->{$code};
    $logkey = "" unless defined $logkey;

    # duplicate prevention for most keys
    return if ($diff_ms < 300_000
               && $logkey ne "move_up" && $logkey ne "move_down"
               && $logkey ne "page_up" && $logkey ne "page_down");


    if ($logkey =~ /^\d$/ && $logkey) {  # number 2-9
	addNumber($logkey);
	return;
    }

    if ($logkey eq "select") {
	my $cur = $list->currentItem();
	return if $cur == -1;
	my $li = $list->item($cur);  # list item
	my $type = $li->type();
	my $file = $li->filename();
	return unless $file;

	# open/close folder
	if ($type eq "diropen" || $type eq "dirclosed") {
	    my $dir = $dir->{$file};
	    $dir->{'open'} = ! $dir->{'open'};
	    updateListWidget();
	    $list->setSelected($cur, 1);
	    return;
	}

	# play music
	if ($type eq "song" || $type eq "playlist") {
	    system("xmms", "/mp3/$file");
	    return;
	}

	return;
    }

    if ($logkey eq "move_down" || $logkey eq "page_down") {
	setHelp(0);
	my $cur = $list->currentItem();
	my $num = $list->count();
	return unless $num;
	my $new = $cur + ($logkey eq "move_down" ? 1 : ($list->numItemsVisible()-1));
	$new = $num-1 if $new >= $num;
	$list->setCurrentItem($new);
	return;
    }

    if ($logkey eq "move_up" || $logkey eq "page_up") {
	setHelp(0);
	my $cur = $list->currentItem();
	my $num = $list->count();
	return unless $num;
	my $new = $cur - ($logkey eq "move_up" ? 1 : ($list->numItemsVisible()-1));
        $new = 0 if $new < 0;
	$list->setCurrentItem($new);
	return;
    }

    if ($logkey eq "status") {
	setHelp(! $help_on);
	return;
    }

    if ($logkey eq "fast_forward") { system("xmms", "--fwd"); return ; }
    if ($logkey eq "rewind") { system("xmms", "--rew"); return ; }
    if ($logkey eq "stop") { system("xmms", "--stop"); return ; }
    if ($logkey eq "pause") { system("xmms", "--pause"); return ; }
    if ($logkey eq "play") { system("xmms", "--play"); return ; }

    if ($logkey eq "clear_one") { clearOne(); return;  }
    if ($logkey eq "clear_all") { clearAll(); return;  }

    print STDERR "code: $code, logkey: $logkey\n";
}

1;
